#include <u.h>
#include <libc.h>
#include <bio.h>

/*
	Copyright 2023
	
	The author does not remember how this program works, or if it even does.
	
	Permission to use, copy, modify, and distribute this software for any purpose with or without fee is hereby granted, provided information regarding the author is erased and not disclosed to any third party.
	
	THE SOFTWARE IS PROVIDED "AS IS". THE AUTHOR DISCLAIMS ALL WARRANTIES INCLUDING ALL IMPLIED WARRANTIES. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR DAMAGES.
*/

/* "accessing uninitialized data (for example, popping an empty stack or reading an uninitialized variable) is undefined behaviour." */

/* TODO: check if things error <- TODO: what was that in reference to */
/* TODO: nesting dosen't work for everything */

union var
{
	s32int i;
	union var *reg;
	char *str;
};
#define STACKSIZE 1024
union var stack[STACKSIZE];
usize top = -1;

struct frame
{
	char *ins;
	int incond;
	char *body;
	char *cond;
};
#define CALLSIZE 256
struct frame callstack[CALLSIZE];
usize calltop = -1;

union var regs['z'-'a'+1];

Biobuf prog;
Biobuf in;
Biobuf out;


void
usage(void)
{
	print("oops\n");
	exits("usage");
}

/* leaving this here because I was proud of it, unfortunately pipes & stdin. */
/* untested */
/*
char *
readquote(void)
{
	char *start = Boffset(&prog);
	int len;
	char *s;
	
	for(int nest = 1; nest > 0)
		switch(Bread(&prog))
		{
		case '[':
			nest++;
			break;
		case ']':
			nest--;
			break;
		case Beof:
			exits("eof in quote");
		default:
			break;
		}
	
	len = Boffset(&prog)-start;
	s = malloc(len)
	Bseek(&prog, start, 0);
	Bread(&prog, s, len);
	s[len-1] = 0;
	return s;
}
*/

int
getins(void)
{
	if(calltop == -1)
		return Bgetc(&prog);
	else
		return *(callstack[calltop].ins++);
}

void
ungetins(void)
{
	if(calltop == -1)
		Bungetc(&prog);
	else
		callstack[calltop].ins--;
}

char *
readquote(void)
{
	char *s = nil;
	int len = 0;
	
	for(int nest = 1; nest > 0;)
	{
		s = realloc(s, ++len); /* this is awful - change it later */
		switch(s[len-1] = getins())
		{
		case '[':
			nest++;
			break;
		case ']':
			nest--;
			break;
		case 0:
			exits("null char in quote");
		case Beof:
			exits("eof in quote");
		default:
			break;
		}
	}
	
	s[len-1] = 0;
	return s;
}

void
main(int argc, char *argv[])
{
	Binit(&in, 0, OREAD);
	Binit(&out, 1, OWRITE);
	
	if(argc <= 1 || (argv[1][0] == '-' && argv[1][1] == 0))
		Binit(&prog, 0, OREAD);
	else if(argc == 2)
	{
		int fd = open(argv[1], OREAD);
		Binit(&prog, fd, OREAD);
	}
	else
		usage();
	
	/* Ignore Shebang */
	if(Bgetc(&prog) == '#')
		for(int c = Bgetc(&prog); c != Beof && c != '\n'; c = Bgetc(&prog));
	else
		Bungetc(&prog);
	
	for(int c = getins(); c != Beof; c = getins())
	{
		union var x,y,z;
		
	//	Bprint(&out, "%c", c);
	//	Bflush(&out);
		
		switch(c)
		{
			/* Comments */
			case '{':
				for(c = getins(); c != Beof && c != '}'; c = getins());
				break;
			
			/* Print Message */
			case '"':
				for(c = getins(); c != Beof && c != '"'; c = getins())
					Bputc(&out, c);
				break;
			
			/* Reg Access */
			case ';':
				stack[top] = *(stack[top].reg);
				break;
			case ':':
				*(stack[top].reg) = stack[top-1];
				top -= 2;
				break;
			
			/* Stack */
			case '$':
				stack[top+1] = stack[top];
				top++;
				break;
			case '%':
				top--;
				break;
			case '\\':
				x = stack[top];
				stack[top] = stack[top-1];
				stack[top-1] = x;
				break;
			case '@':
				x = stack[top-2];
				y = stack[top-1];
				z = stack[top  ];
				stack[top-2] = y;
				stack[top-1] = z;
				stack[top  ] = x;
				break;
			case 'ø':
			case 'O':
				stack[top] = stack[top-1-stack[top].i];
				break;
			
			/* Number Ops */
			case '+':
				x = stack[top--];
				stack[top].i += x.i;
				break;
			case '-':
				x = stack[top--];
				stack[top].i -= x.i;
				break;
			case '*':
				x = stack[top--];
				stack[top].i *= x.i;
				break;
			case '/':
				x = stack[top--];
				stack[top].i /= x.i;
				break;
			case '_':
				stack[top].i = -stack[top].i;
				break;
			
			/* Bit Ops */
			case '&':
				x = stack[top--];
				stack[top].i &= x.i;
				break;
			case '|':
				x = stack[top--];
				stack[top].i |= x.i;
				break;
			case '~':
				stack[top].i = ~stack[top].i;
				break;
			
			/* Queries */
			case '=':
				x = stack[top--];
				stack[top].i = -(stack[top].i == x.i);
				break;
			case '>':
				x = stack[top--];
				stack[top].i = -(stack[top].i > x.i);
				break;
			case '<':
				x = stack[top--];
				stack[top].i = -(stack[top].i < x.i);
				break;
			
			/* Calls */
			case '!':
				calltop++;
				callstack[calltop].body = stack[top].str;
				callstack[calltop].cond = nil;
				callstack[calltop].ins = stack[top].str;
				callstack[calltop].incond = 0;
				top--;
				break;
			case '?':
				if(stack[top-1].i != 0)
				{
					calltop++;
					callstack[calltop].body = stack[top].str;
					callstack[calltop].cond = nil;
					callstack[calltop].ins = stack[top].str;
					callstack[calltop].incond = 0;
				}
				top -= 2;
				break;
			case '#':
				calltop++;
				callstack[calltop].body = stack[top].str;
				callstack[calltop].cond = stack[top-1].str;
				callstack[calltop].ins = stack[top].str;
				callstack[calltop].incond = 0;
				top -= 2;
				break;
			
			/* Returns */
			case 0:
				// FIXME: memleak of strings, can't free as may ref'd elsewhere
				if(callstack[calltop].incond)
				{
					if(stack[top--].i != 0)
					{
						callstack[calltop].incond = 0;
						callstack[calltop].ins = callstack[calltop].body;
					}
					else
						calltop--;
				}
				else if(callstack[calltop].cond != nil)
				{
					callstack[calltop].incond = 1;
					callstack[calltop].ins = callstack[calltop].cond;
				}
				else
					calltop--;
				break;
			
			/* Quote */
			case '[':
				stack[++top].str = readquote();
				break;
			case ']':
				// pass
				break;
			
			/* IO */
			case '^':
				c = Bgetc(&in);
				stack[++top].i = (c == Beof ? -1 : c);
				break;
			case ',':
				Bputc(&out, stack[top--].i);
				break;
			case '.':
				Bprint(&out, "%d", stack[top--].i);
				break;
			case 'ß':
			case 'B':
				Bflush(&out);
				break;
			
			/* Character Literal */
			case '\'':
				x.i = getins();
				if(x.i != 0 && x.i != Beof)
					stack[++top].i = x.i;
				else
					ungetins();
				break;
			
			/* Integer Literal */
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
				stack[++top].i = c - '0';
				for(c = getins(); '0' <= c && c <= '9'; c = getins())
					stack[top].i = stack[top].i*10 + c - '0';
				ungetins();
				break;
			
			default:
				/* Register Ref */
				if('a' <= c && c <= 'z')
					stack[++top].reg = &(regs[c-'a']);
				break;
		}
		
	//	Bflush(&out);
	}
	
	Bflush(&out);
	exits(0);
}
